# SPDX-FileCopyrightText: ☭ Emery Hemingway
# SPDX-License-Identifier: Unlicense

import std/[deques, streams]
import bumpy, pixie, sdl2

type
  Rect = bumpy.Rect

  Pane* = ref object
    spans: seq[Span]
    arrangement: Arrangement
    image: Image
    texture: TexturePtr

  Well* = ref object
    panes: Deque[Pane]
    width, height: int

using
  pane: Pane
  well: Well

proc newWell*(width, height: int): Well =
  result = Well(width: width, height: height)
  result.panes.addFirst Pane()

proc dimensions(well): Vec2 = vec2(float well.width, float well.height)

proc margins*(well): (int, int) = (well.width div 9, well.height div 9)

proc rect*(well): Rect =
  result.wh = well.dimensions

proc rectAt(well; i: int): Rect =
  if i < well.panes.len:
    let
      (quo, rem) = divmod(i, 3)
      shiftOff = quo.succ
      scaleDiv = 1 shl quo.succ
      fullWh = well.dimensions
    result.w = float(well.width shr shiftOff)
    result.h = float(well.height shr shiftOff)
    case rem
    of 0:
      result.xy = fullWh - (result.wh * 2.0)
    of 1:
      result.x = fullWh.x - result.w
      result.y = fullWh.y - (result.h * 2.0)
    of 2:
      result.x = fullWh.x - (result.w * 2.0)
      result.y = fullWh.y - result.h
    else: discard

proc place*(well; offset, width, height: int): (Pane, Rect) =
  ## Return the `Pane` at `offset` from the top of `well` or `nil` if
  ## the offset is too deep. The position and size of the `Pane` is
  ## returned as well.
  if offset < well.panes.len:
    result[0] = well.panes[offset]
    result[1] = well.rectAt(offset)

proc append*(well; text: string; font: Font) =
  assert well.panes.len > 0
  let span = newSpan(text & "\n", font)
  while true:
    let pane = well.panes.peekLast()
    pane.spans.add(span)
    var
      arrangement = typeset(pane.spans, well.dimensions)
      bounds = layoutBounds arrangement
    if bounds.y <= well.dimensions.y:
      doAssert pane.spans.len > 0, "text does not find on a single pane - " & $bounds & $well.dimensions
      pane.arrangement = arrangement
      break
    else:
      discard pane.spans.pop()
      well.panes.addLast Pane()

proc append*(well; stream: Stream; font: Font) =
  var line: string
  while readLine(stream, line):
    append(well, line, font)

#[
proc render*(well; index: Natural; scale: float): Image =
  if index < well.panes.len:
    let dimensions = well.dimensions * scale
    if dimensions.x > 8 and dimensions.y > 8:
      result = newImage(int dimensions.x, int dimensions.y)
      fill(result, rgba(255, 255, 255, 255))
      fillText(well.panes[index].arrangment, translate(well.margins))
]#

type Intersection = tuple
  src, dst: Rect
  index: int

const
  amask = uint32 0xff000000
  rmask = uint32 0x000000ff
  gmask = uint32 0x0000ff00
  bmask = uint32 0x00ff0000

proc texture*(well; index: int; renderer: RendererPtr): TexturePtr =
  if index < well.panes.len:
    let pane = well.panes[index]
    if pane.texture.isNil and not pane.arrangement.isNil:
      assert pane.image.isNil
      let
        w = (well.width div 9) * 7
        h = (well.height div 9) * 7
      pane.image = newImage(w, h)
      pane.image.fill(rgba(255, 255, 255, 255))
      pane.image.fillText(pane.arrangement)
      var surface = createRGBSurfaceFrom(
          pane.image.data[0].addr,
          pane.image.width.cint,
          pane.image.height.cint,
          cint 32,
          pane.image.width.cint shl 2,
          rmask, gmask, bmask, amask,
        )
      assert(not surface.isNil, $getError())
      pane.texture = createTextureFromSurface(renderer, surface)
      destroy(surface)
      assert(not pane.texture.isNil, $getError())
    result = pane.texture

iterator intersectingPanes*(well; view: Rect): Intersection =
  var
    intersect: Intersection
    zoom = 1.0
  while intersect.index < well.panes.len:
    var paneRect = well.rectAt(intersect.index)
    if (intersect.index mod 3) == 0:
      let quad = rect(paneRect.xy, paneRect.wh * 2.0)
      if not view.overlaps quad:
        # all further panes are non-intersecting
        break
      zoom = zoom * 2.0
    let margin = paneRect.wh / 9.0
    paneRect.xy = paneRect.xy + margin
    paneRect.wh = paneRect.wh - (margin * 2.0)
    if view.overlaps paneRect:
      intersect.dst = view and paneRect
      intersect.src = rect(intersect.dst.xy - paneRect.xy, intersect.dst.wh) * zoom
      yield(intersect)
    inc(intersect.index)
